/* Kishimot's Menu ver 0.06 */
#include <ncurses.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define YSIZE 28
#define XSIZE 76

#define PgUP 1
#define PgDOWN 2
#define CURUP 3
#define CURDOWN 4
#define CURLEFT 5
#define CURRIGHT 6
#define F1 7
#define F2 8
#define F3 9

typedef struct{
	char *filename;
	char *string;
} Item;

Item *item;
int itemnum=0;
int topnum=0;
char baseDir[BUFSIZ];
char *kmenuTitle;

int
mygetch(){
	char c;
	c = getch();
	if(c == ''){
		c = getch();
		if(c == '['){
			c = getch();
			if(c == 'A') c = CURUP;
			else if(c == 'B') c = CURDOWN;
			else if(c == 'C') c = CURRIGHT;
			else if(c == 'D') c = CURLEFT;
			else if(c == '5') {
				c = getch();
				if(c == '~') c=PgUP;
			}else if(c == '6') {
				c = getch();
				if(c == '~') c=PgDOWN;
			} else if(c == '[') {
				c = getch();
				if(c == 'A') c=F1;
				if(c == 'B') c=F2;
				if(c == 'C') c=F3;
			}
		}
	}
	return c;
}

char *
getLine(char *buf){
	int i;

	i=0;
	buf[strlen(buf)-1]='\0';
	while(1){
		if(buf[i]=='\0') break;
		if (((unsigned char)buf[i] & 0x80) != 0) i++;
		i++;
		if(i>XSIZE) {
			buf[i]='\0';
			break;
		}
	}
	return buf;
}

void
crLock(){
	char lockf[256];
	FILE *fp;

	sprintf(lockf,"%s/.kmenuLock",getenv("HOME"));
	if(NULL != (fp=fopen(lockf,"r"))){
		fprintf(stderr,"~/.kmenuLock already exists.\n");
		exit(1);
	}
	creat(lockf,S_IRWXU);
}

void
delLock(){
	char lockf[256];

	sprintf(lockf,"%s/.kmenuLock",getenv("HOME"));
	remove(lockf);
}


char *
readFile(char *file){
	FILE *fp;
	char buf[BUFSIZ],fn[BUFSIZ];

	sprintf(fn,"%s/%s",baseDir,file);
	if((fp=fopen(fn,"r"))==NULL){
		fprintf(stderr,"can't open %s\n",fn);
		fclose(fp);
		delLock();
		exit(1);
	}

	/* dummy */
	fgets(buf,BUFSIZ,fp);
	fgets(buf,BUFSIZ,fp);
	fclose(fp);
	return strdup(getLine(buf+1));
}

void
readTitle(){
	FILE *fp;
	char buf[BUFSIZ],fn[BUFSIZ];

	sprintf(fn,"%s/Title",baseDir);
	if((fp=fopen(fn,"r"))==NULL){
		fprintf(stderr,"can't open title\n",fn);
		fclose(fp);
		delLock();
		exit(1);
	}

	fgets(buf,BUFSIZ,fp);
	fclose(fp);
	kmenuTitle = strdup(getLine(buf));
}

void
getMenuItem(){
	DIR *dp;
	struct direct *dir;

	item = (Item *)malloc(sizeof(Item)*(2));

	if((dp=opendir(baseDir))==NULL){
		fprintf(stderr,"can't open %s\n",baseDir);
		closedir(dp);
		delLock();
		exit(1);
	}

	while((dir = readdir(dp))!= NULL){
		if(dir->d_ino ==0) continue;
		if(strcmp(dir->d_name,"..")==0) continue;
		if(strcmp(dir->d_name,".")==0) continue;
		if(strcmp(dir->d_name,"Title")==0) {
			readTitle();
			continue;
		}

		item = (Item *)realloc(item,(sizeof(Item)*(itemnum+1)));

		item[itemnum].filename = strdup(dir->d_name);
		item[itemnum].string = readFile(dir->d_name);
		itemnum++;
	}
	closedir(dp);
}

void
winInit(){
	initscr();
	cbreak(); noecho(); nonl();
	clear();
}

void
showList(){
	int i;
	clear();
	for(i=0;i<YSIZE;i++){
		if(i+topnum>=itemnum) break;
		move(0,0);
		standout();
		printw("%-80s\n",kmenuTitle);
		standend();
		move(i+1,0);
		printw("  %s",item[topnum+i].string);
	}
	refresh();
}

void
prompt(int num){
	move(num+1,0);
	printw(">>");
	refresh();
}

void
deprompt(int num){
	move(num+1,0);
	printw("  ");
	refresh();
}

void
doIt(int num){
	char sh[BUFSIZ];
	sprintf(sh,"%s/%s",baseDir,item[topnum+num].filename);
	nocbreak(); echo(); nl();
	system(sh);
	cbreak(); noecho(); nonl();
	clear();
	showList();
	prompt(num);
}

void
eventLoop(){
	char c;
	int tn;

	showList();
	prompt(0);
	tn=0;
	while('Q' != (c = mygetch())){
		if((c=='j')||(c=='z')||(c==CURDOWN)||(c==F2)){
			if(tn<itemnum-topnum-1){
				if(tn<YSIZE-1){
					deprompt(tn);
					tn++;
					prompt(tn);
				}else{
					if(topnum+YSIZE<itemnum){
						topnum+=YSIZE;
						showList();
						tn=0;
						prompt(tn);
					}
				}
			}
		}else if((c=='k')||(c=='a')||(c==CURUP)||(c==F1)){
			if(tn>0){
				deprompt(tn);
				tn--;
				prompt(tn);
			}else{
				if(topnum>0){
					topnum-=YSIZE;
					if(topnum<0){
						topnum=0;
					}
					showList();
					tn=YSIZE-1;
					prompt(tn);
				}
			}
		}else if((c=='n')||(c==']')||(c==PgDOWN)){
			if(topnum+YSIZE<itemnum){
				topnum+=YSIZE;
				showList();
				tn=0;
				prompt(tn);
			}
		}else if((c=='p')||(c=='[')||(c==PgUP)){
			if(topnum>0){
				topnum-=YSIZE;
				if(topnum<0){
					topnum=0;
				}
				showList();
				tn=YSIZE-1;
				prompt(tn);
			}
		}else if((c=='\r')||(c==' ')||(c==F3)){
			doIt(tn);
		}else if(c=='l'){
			showList();
			prompt(tn);
		}
	}
}

void
winEnd(){
	nocbreak(); echo(); nl();
	endwin();
}

int
main(){
	char *bdir;

	if(NULL != (bdir = getenv("KMENUDIR"))){
		strcpy(baseDir,bdir);
	}else{
		sprintf(baseDir,"%s/.kmenu/",getenv("HOME"));
	}
	kmenuTitle= "                            's ˤ                            ";
	crLock();
	getMenuItem();
	winInit();
	eventLoop();
	winEnd();
	delLock();
}


